xquery version "3.1";

(:~
 : This module transforms TEI to XHTML according to the needs of the Ahiqar project.
 : 
 : There are four types of outcomes of the transformation:
 :  * some elements are ignored ($tei2html:ignored-elements),
 :  * some elements are transformed to xhtml:div ($tei2html:block-elements),
 :  * elements that reference another element (indicated by @target, @next, and
 :    @prev) are transformed to xhtml:a,
 :  * the rest of the elements are transformed to xhtml:span.
 : 
 : In any case the resulting XHTML element hold the original TEI element's name
 : in its class attribute as well as all attributes.
 : The id attribute, which has been added automatically by the TextAPI and holds 
 : the TEI node number, is preserved in any case.
 : 
 : @see https://intranet.sub.uni-goettingen.de/display/prjAhiqar/Text+Styling+Specification
 :)

module namespace tei2html="http://ahikar.sub.uni-goettingen.de/ns/tei2html";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xhtml="http://www.w3.org/1999/xhtml";

import module namespace tei2html-motifs="http://ahikar.sub.uni-goettingen.de/ns/tei2html/motifs" at "tei2html-motifs.xqm";
import module namespace tei2html-text="http://ahikar.sub.uni-goettingen.de/ns/tei2html/textprocessing" at "tei2html-textprocessing.xqm";

declare variable $tei2html:block-elements := ("ab", "body", "cb", "l", "lg");
declare variable $tei2html:ignored-elements := ("lb", "milestone", "pb");


declare function tei2html:transform($nodes as node()*)
as node()* {
    for $node in $nodes return
        typeswitch ($node)
        case text() return
            tei2html-text:process-text($node)
            
        case processing-instruction() return
            ()
                
        case element(tei:body) return
            if (tei2html:has-page-content($node)) then
                tei2html:make-default-return($node)
            else
                tei2html:make-vacant-page()
            
        case element(tei:head) return
            element xhtml:h1 {
                tei2html:make-id-and-class-attributes($node),
                tei2html:transform($node/node())
            }
            
        case element(tei:w) return
            (
                element xhtml:span {
                    attribute id {$node/@xml:id},
                    attribute class {"token"},
                    tei2html:transform($node/node())
                },
                text{" "}
            )
            
        case element(tei:ab) return
            let $desc-comment-start := $node/descendant::node()[self::processing-instruction("oxy_comment_start")]
            let $desc-comment-end := $node/descendant::node()[self::processing-instruction("oxy_comment_end")]
            return
                if ($desc-comment-start
                or tei2html-motifs:is-line-in-motif($node)) then
                    (: one or more nested motifs start and end in the same line :)
                    if(count($desc-comment-start) gt 1
                    and count($desc-comment-start) = count($desc-comment-end)) then
                        let $nodes-before-first-motif := tei2html-motifs:get-nodes-before-pi($desc-comment-start[1], $node)
                        let $nodes-after-last-motif := $node/child::node()[. >> $desc-comment-end[last()]]
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:transform($nodes-before-first-motif),
                                tei2html:make-several-motifs-start-and-end-in-line($desc-comment-start[1], $node),
                                tei2html:transform($nodes-after-last-motif)
                            }
                            
                    (: one or more nested motifs span over several lines.
                    part 1: beginning of the motif(s) in the same line :)
                    else if (count($desc-comment-start) gt 1
                    and not($desc-comment-end)) then
                        let $nodes-before-first-motif := tei2html-motifs:get-nodes-before-pi($desc-comment-start[1], $node)
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:transform($nodes-before-first-motif),
                                tei2html:make-several-motifs-start-in-same-line($desc-comment-start[1], $node)
                            }
                            
                    (: one or more nested motifs span over several lines.
                    part 2.1: (all) motif(s) continue over a complete line :)
                    else if (count($desc-comment-start) = 0
                    and count($desc-comment-end) = 0
                    and tei2html-motifs:is-part-of-more-than-one-unclosed-motif($node)) then
                        let $previous-unclosed-motifs := tei2html-motifs:get-unclosed-motifs-before-line($node)
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:make-several-motifs-span-complete-line($previous-unclosed-motifs, $node)
                            }
                            
                    (: one or more nested motifs span over several lines.
                    part 2.2: (all) motif(s) continue over a complete line, 
                    one or more new motif(s) start(s) in line :)
                    else if(count($desc-comment-start) ge 1
                    and tei2html-motifs:is-part-of-one-or-more-unclosed-motifs($node)) then
                        let $previous-unclosed-motifs := tei2html-motifs:get-unclosed-motifs-before-line($node)
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:make-several-motifs-span-and-start-in-line($previous-unclosed-motifs, $node)
                            } 
                        
                    (: one or more nested motifs span over several lines.
                    part 3: end of all motif(s) in the same line :)
                    else if (count($desc-comment-end) gt 1
                    and not($desc-comment-start)) then
                        let $nodes-after-last-motif-end := $node/child::node()[. >> $desc-comment-end[last()]]
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:make-several-motifs-end-in-line($desc-comment-end, $node),
                                tei2html:transform($nodes-after-last-motif-end)
                            }
                        
                        
                    (: a single motif starts in a line :)    
                    else if(count($desc-comment-start) = 1
                    and count($desc-comment-end) = 0) then
                        let $nodes-before-motif-start := tei2html-motifs:get-nodes-before-pi($desc-comment-start, $node)

                        let $nodes-in-motif := $node/child::node()[. >> $desc-comment-start] 
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:transform($nodes-before-motif-start),
                                tei2html:make-simple-mark-with-data-next($desc-comment-start, $node, $nodes-in-motif)
                            }
                    
                    (: a single motif is continued in a line :)
                    else if (count($desc-comment-start) = 0
                    and count($desc-comment-end) = 0
                    and tei2html-motifs:is-line-in-motif($node)) then
                        let $previous-motif-start := $node/preceding::node()[self::processing-instruction("oxy_comment_start")][1]
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:make-simple-mark-with-data-next($previous-motif-start, $node, $node/node())
                            }
                    
                    (: a single motif ends in a line :)    
                    else if(count($desc-comment-end) = 1
                    and count($desc-comment-start) = 0) then
                        let $nodes-in-motif := tei2html-motifs:get-nodes-before-pi($desc-comment-end, $node)
                        let $nodes-after-motif-end := $node/child::node()[. >> $desc-comment-end]
                        let $corresp-motif-start := $desc-comment-end/preceding::node()[self::processing-instruction("oxy_comment_start")][1]
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:make-simple-mark($corresp-motif-start, $node, $nodes-in-motif),
                                tei2html:transform($nodes-after-motif-end)
                            }
                            
                    
                    (: a single motif starts and ends in a line :)
                    else if(count($desc-comment-start) = 1
                    and count($desc-comment-start) = count($desc-comment-end)) then
                        let $nodes-before-motif := tei2html-motifs:get-nodes-before-pi($desc-comment-start, $node)
                        let $nodes-after-motif := $node/child::node()[. >> $desc-comment-end]
                        let $nodes-in-motif := $node/child::node()[. >> $desc-comment-start and . << $desc-comment-end]
                        return
                            element div {
                                tei2html:make-id-and-class-attributes($node),
                                tei2html:transform($nodes-before-motif),
                                tei2html:make-simple-mark($desc-comment-start, $node, $nodes-in-motif),
                                tei2html:transform($nodes-after-motif)
                            }
                    
                    else
                        ()
            else
                tei2html:make-default-return($node)
            
        default return
            tei2html:make-default-return($node)
};

declare function tei2html:has-page-content($body as element(tei:body))
as xs:boolean {
    $body/descendant::*
    and $body/descendant::text()[matches(., "[\w]")]
};

declare function tei2html:make-default-return($node as node()*)
as node()* {
    if (tei2html:is-block-element($node)) then
        local:make-xhtml-div($node)
        
    else if (tei2html:references-another-element($node)) then
        tei2html:make-xhtml-a($node)
        
    else if (tei2html:is-ignored-element($node)) then
        ()
        
    else
        local:make-xhtml-span($node)
};

declare function tei2html:is-block-element($node as node()*)
as xs:boolean {
    local-name($node) = $tei2html:block-elements
};

declare function tei2html:is-ignored-element($node as node()*)
as xs:boolean {
    local-name($node) = $tei2html:ignored-elements
    or $node[self::comment()]
};

declare function local:make-xhtml-span($element as element())
as element(xhtml:span) {
    element xhtml:span {
        tei2html:make-id-and-class-attributes($element),
        tei2html:transform($element/node())
    }
};

declare function local:make-xhtml-div($element as element())
as element(xhtml:div) {
    element xhtml:div {
        tei2html:make-id-and-class-attributes($element),
        tei2html:transform($element/node())
    }
};

declare function tei2html:make-class-attribute-values($element as element())
as xs:string {
    let $element-name := $element/local-name()
    let $attribute-values :=
        for $attr in $element/(@* except @id) return
            if ($attr = "color(red)") then
                "red"
            else
                $attr
    return
        string-join(($element-name, $attribute-values), " ")
};

declare function tei2html:make-vacant-page()
as element(xhtml:div) {
    element xhtml:div {
        attribute class {"vacant-page"}
    }
};


declare function tei2html:references-another-element($node as node())
as xs:boolean {
    exists($node/@next)
    or exists($node/@prev)
    or exists($node/@target)
};


declare function tei2html:make-xhtml-a($element as element())
as element(xhtml:a) {
    element xhtml:a {
        attribute id {$element/@id},
        attribute class {$element/local-name()},
        attribute href {tei2html:get-href-value($element)},
        tei2html:transform($element/node())
    }
};


(: While references work on basis of @xml:id (e.g. 'seg_1') in TEI, we link to
 : the TEI node's ID that has been generated by generate-id() instead in XHTML.
 : The reason for this is that we also use the node ID for the AnnotationAPI so
 : the data we receive at this point looks like this:
 : 
 : <seg id="N.1.1.1.2.3" type="colophon">...</seg>
 : 
 : Since @id has to preserved for the AnnotationAPI, we cannot simply store
 : @xml:id in @id.
 :)
declare function tei2html:get-href-value($element as element())
as xs:string? {
    let $xmlid := tei2html:get-referenced-xmlid($element)
    let $referenced-id := tei2html:find-referenced-node-id($element, $xmlid)
    return
        "#" || $referenced-id
};

declare function tei2html:find-referenced-node-id($element as element(),
    $xmlid as xs:string)
as xs:string? {
    $element/root()//*[@xml:id = $xmlid]/@id
};

declare function tei2html:get-referenced-xmlid($element as element())
as xs:string? {
    if ($element/@next) then
        $element/@next => substring-after("#")
    else if ($element/@prev) then
        $element/@prev => substring-after("#")
    else if ($element/@target) then
        tokenize($element/@target, " ")[1]
        => substring-after("#")
    else
        ()
};

(:~
 : This function handles nested motifs which all start and begin in the same line.
 : They have a structure similar to
 : 
 : ...{..[....(..).]...........}....
 : 
 : where the dots represent words and the braces the start/beginning of a motif.
 : The function starts to process the line from the outmost motif and recursively
 : progresses to the middle of the line until the innermost motif is reached in
 : the base case.
 : 
 : @param $motif-start The start comment of the current motif.
 : @param $node The current line to be handled.
 : @retun mark An HTML representing the current motif.
 :)
declare function tei2html:make-several-motifs-start-and-end-in-line($motif-start as processing-instruction(),
    $node as element(tei:ab))
as element(mark) {
    let $next-motif-start := $motif-start/following::processing-instruction("oxy_comment_start")[ancestor::tei:ab = $motif-start/ancestor::tei:ab][1]
    let $motif-end := tei2html-motifs:get-corresponding-end-of-motif($motif-start, $node)
    
    return
        (: recursive step: going from the outermost motif further to the middle :)
        if ($next-motif-start) then
            let $next-motif-end := tei2html-motifs:get-corresponding-end-of-motif($next-motif-start, $node)
            let $nodes-before-next-motif-start := $node/child::node()[. >> $motif-start and . << $next-motif-start]
            let $nodes-after-next-motif-end := $node/child::node()[. >> $next-motif-end and . << $motif-end]
            return
                element mark {
                    tei2html-motifs:make-attributes-for-mark($motif-start, $node),
                    tei2html:transform($nodes-before-next-motif-start),
                    tei2html:make-several-motifs-start-and-end-in-line($next-motif-start, $node),
                    tei2html:transform($nodes-after-next-motif-end)
                }
        (: base case: the innermost motif :)
        else
            let $nodes-in-motif := $node/child::node()[. >> $motif-start and . << $motif-end]
            return
                tei2html:make-simple-mark($motif-start, $node, $nodes-in-motif)
};




(:~
 : This function handles nested motifs which span the complete line while one or
 : more new motifs start in this line.
 : The line has a structure similar to
 : 
 : ....(...[..............
 : > ...{.................
 : ....}....])............
 : 
 : where the dots represent words and the braces the start/beginning of a motif.
 : The arrow marks the line in focus.
 : 
 : The whole phenomenon is covered by two functions, one of them being this one
 : while the other is make-motifs-start-in-line-spanned-by-one-or-more-motifs.
 : This function deals with setting the motifs that have begun in one of the
 : previous lines as well as the text before the newly started motif(s).
 : 
 : @param $previous-unclosed-motifs A sequence of all the start comments that haven't been closed yet.
 : @param $node The current line to be handled.
 : @retun mark An HTML representing the current motif.
 :)
declare function tei2html:make-several-motifs-span-and-start-in-line($previous-unclosed-motifs as processing-instruction()*,
    $node as element(tei:ab))
as element(mark) {
    let $first-motif-start-in-line := $node/descendant::processing-instruction("oxy_comment_start")[1]
    let $nodes-before-first-motif-start := tei2html-motifs:get-nodes-before-pi($first-motif-start-in-line, $node)
    return
        (: recursive step: set a mark for each unclosed comment that encompasses
        the complete line :)
        if (count($previous-unclosed-motifs) ge 1) then
            let $current-motif := $previous-unclosed-motifs[1]
            return
                element mark {
                    tei2html:make-attributes-with-data-next-for-mark($current-motif, $node),
                    tei2html:transform($nodes-before-first-motif-start),
                    tei2html:make-several-motifs-span-and-start-in-line($previous-unclosed-motifs[position() gt 1], $node)
                }
        (: base step: all of the unclosed motifs have been handled, now the motifs
        that are started in this line are processed :)
        else
            tei2html:make-motifs-start-in-line-spanned-by-one-or-more-motifs($first-motif-start-in-line, $node)
};

(:~
 : This function handles nested motifs which span the complete line while one or
 : more new motifs start in this line.
 : The line has a structure similar to
 : 
 : ....(...[..............
 : > ...{.................
 : ....}....])............
 : 
 : where the dots represent words and the braces the start/beginning of a motif.
 : The arrow marks the line in focus.
 : 
 : The whole phenomenon is covered by two functions, one of them being this one
 : while the other is make-several-motifs-span-and-start-in-line.
 : This function deals with setting the motifs that begin in the current line.
 : They are processed from left to right.
 : 
 : @param $motif-start The start comment of the current motif.
 : @param $node The current line to be handled.
 : @retun mark An HTML representing the current motif.
 :)
declare function tei2html:make-motifs-start-in-line-spanned-by-one-or-more-motifs($motif-start as processing-instruction(),
    $node as element(tei:ab))
as element(mark) {
    let $next-motif-start := $motif-start/following::processing-instruction("oxy_comment_start")[ancestor::tei:ab = $motif-start/ancestor::tei:ab][1]
    return
        (: recursive step: traverse the start comments of the motifs from left
        to right :)
        if ($next-motif-start) then
            let $nodes-before-next-motif-start := $node/child::node()[. >> $motif-start and . << $next-motif-start]
            
            return
                element mark {
                    tei2html:make-attributes-with-data-next-for-mark($motif-start, $node),
                    tei2html:transform($nodes-before-next-motif-start),
                    tei2html:make-motifs-start-in-line-spanned-by-one-or-more-motifs($next-motif-start, $node)
                }
        (: base case: the rightmost start comment has been reached :)
        else
            let $nodes-in-motif := $node/child::node()[. >> $motif-start]
            return
                tei2html:make-simple-mark-with-data-next($motif-start, $node, $nodes-in-motif)
};

(:~
 : This function handles nested motifs which all start in the same line.
 : They continue and/or end in one og the next lines.
 : They have a structure similar to
 : 
 : > ..(...[..............
 : .......................
 : .......]...........)...
 : 
 : where the dots represent words and the braces the start/beginning of a motif.
 : The arrow marks the line in focus
 : 
 : The function starts to process the line from the leftmost start comment and 
 : recursively progresses to the end of the line.
 : 
 : @param $motif-start The start comment of the current motif.
 : @param $node The current line to be handled.
 : @retun mark An HTML representing the current motif.
 :)
declare function tei2html:make-several-motifs-start-in-same-line($motif-start as processing-instruction(),
    $node as element(tei:ab))
as element(mark) {
    let $next-motif-start := $motif-start/following::processing-instruction("oxy_comment_start")[ancestor::tei:ab = $motif-start/ancestor::tei:ab][1]
    return
        (: recursive step: going from the leftmost motif to the next one :)
        if ($next-motif-start) then
            let $nodes-before-next-motif-start := $node/child::node()[. >> $motif-start and . << $next-motif-start]
            
            return
                element mark {
                    tei2html:make-attributes-with-data-next-for-mark($motif-start, $node),
                    tei2html:transform($nodes-before-next-motif-start),
                    tei2html:make-several-motifs-start-in-same-line($next-motif-start, $node)
                }
        (: base case: the last (= rightmost) motif in the line :)
        else
            let $nodes-in-motif := $node/child::node()[. >> $motif-start]
            return
                tei2html:make-simple-mark-with-data-next($motif-start, $node, $nodes-in-motif)
};

(:~
 : This function handles nested motifs which span the complete line while having
 : started in some line before.
 : The line has a structure similar to
 : 
 : ....(...[..............
 : > .....................
 : .......]...........)...
 : 
 : where the dots represent words and the braces the start/beginning of a motif.
 : The arrow marks the line in focus.
 : 
 : The whole phenomenon is covered by three functions, one of them being this one
 : while the other two are make-several-motifs-start-in-same-line and 
 : make-several-motifs-end-in-line.
 : This function gathers all unclosed motifs and sets a line spanning mark for
 : each one.
 : 
 : @param $unclosed-motif-starts A sequence of start comments of unclosed motifs.
 : @param $node The current line to be handled.
 : @retun mark An HTML representing the current motif.
 :)
declare function tei2html:make-several-motifs-span-complete-line($unclosed-motif-starts as processing-instruction()+,
    $node as element(tei:ab))
as element(mark) {
    (: recursive step: set a mark for each unclosed motif :)
    if (count($unclosed-motif-starts) gt 1) then
        let $current-motif := $unclosed-motif-starts[1]
        return
            element mark {
                tei2html:make-attributes-with-data-next-for-mark($current-motif, $node),
                tei2html:make-several-motifs-span-complete-line($unclosed-motif-starts[position() gt 1], $node)
            }
    (: base step: having reached the last unclosed motif :)
    else
        tei2html:make-simple-mark-with-data-next($unclosed-motif-starts, $node, $node/node())
};

(:~
 : This function handles nested motifs which end in the current line while having
 : started in one of the lines before.
 : The line has a structure similar to
 : 
 : ....(...[..............
 : .......................
 : > .....]...........)...
 : 
 : where the dots represent words and the braces the start/beginning of a motif.
 : The arrow marks the line in focus.
 : 
 : The whole phenomenon is covered by three functions, one of them being this one
 : while the other two are make-several-motifs-start-in-same-line and 
 : make-several-motifs-span-complete-line.
 : This function starts at the rightmost end comment and recursively proceeds to
 : the leftmost end comment of a motif.
 : 
 : @param $motif-ends A sequence of end comments in the line.
 : @param $node The current line to be handled.
 : @retun mark An HTML representing the current motif.
 :)
declare function tei2html:make-several-motifs-end-in-line($motif-ends as processing-instruction()+,
    $node as element(tei:ab))
as element(mark) {
    (: recursive step: process the rightmost motif end and then proceed to the
    next one to the left :)
    if (count($motif-ends) gt 1) then
        let $no-of-items-in-sequence := count($motif-ends)
        let $current-motif-end := $motif-ends[last()]
        let $nodes-until-prev-motif-end := $node/child::node()[. << $current-motif-end and . >> $motif-ends[position() lt $no-of-items-in-sequence]]
        let $corresponding-motif-start := tei2html-motifs:get-corresponding-motif-start($current-motif-end, $node)
        return
            element mark {
                tei2html-motifs:make-attributes-for-mark($corresponding-motif-start, $node),
                tei2html:make-several-motifs-end-in-line($motif-ends[position() lt $no-of-items-in-sequence], $node),
                tei2html:transform($nodes-until-prev-motif-end)
            }
    (: base step: having reached the leftmost end comment :)
    else
        let $corresponding-motif-start := tei2html-motifs:get-corresponding-motif-start($motif-ends, $node)
        let $nodes-before-motif-end := tei2html-motifs:get-nodes-before-pi($motif-ends, $node)
        return
            tei2html:make-simple-mark($corresponding-motif-start, $node, $nodes-before-motif-end)
};

declare function tei2html:make-simple-mark-with-data-next($motif-start as processing-instruction(),
    $node as element(tei:ab),
    $nodes-to-be-transformed as node()+)
as element(tei:mark) {
    element mark {
        tei2html:make-attributes-with-data-next-for-mark($motif-start, $node),
        tei2html:transform($nodes-to-be-transformed)
    }
};

declare function tei2html:make-simple-mark($motif-start as processing-instruction(),
    $node as element(tei:ab),
    $nodes-to-be-transformed as node()+)
as element(tei:mark) {
    element mark {
        tei2html-motifs:make-attributes-for-mark($motif-start, $node),
        tei2html:transform($nodes-to-be-transformed)
    }
};

declare function tei2html:make-id-and-class-attributes($node as element(*))
as attribute()+ {
    attribute id {$node/@id},
    attribute class {
        tei2html:make-class-attribute-values($node)
    }
};

declare function tei2html:make-attributes-with-data-next-for-mark($current-motif as processing-instruction(),
    $node as element(tei:ab))
as attribute()+ {
    tei2html-motifs:make-attributes-for-mark($current-motif, $node),
    tei2html-motifs:make-data-next-attribute($current-motif, $node)
};
